@@ -1,6 +1,6 @@
 /*
    3APA3A simpliest proxy server
-   (c) 2002-2016 by Vladimir Dubrovin <3proxy@3proxy.ru>
+   (c) 2002-2021 by Vladimir Dubrovin <3proxy@3proxy.org>
 
    please read License Agreement
 
@@ -61,9 +61,9 @@ unsigned char * strings[] = {
 	"is coded right now. What you see is a part of work that is done\n"
 	"already.\n"
 	"<p>Please send all your comments to\n"
-	"<A HREF=\"mailto:3proxy@security.nnov.ru\">3proxy@security.nnov.ru</A>\n"
+	"<A HREF=\"mailto:3proxy@3proxy.org\">3proxy@3proxy.org</A>\n"
 	"<p>Documentation:\n"
-	"<A HREF=\"http://3proxy.ru/\">http://3proxy.ru/</A>\n"
+	"<A HREF=\"https://3proxy.org/\">http://3proxy.org/</A>\n"
 	"</tr></td></table>",
 /* 36 */	NULL,
 /* 37 */	NULL,
