@@ -1,6 +1,6 @@
 /*
    3APA3A simpliest proxy server
-   (c) 2002-2016 by Vladimir Dubrovin <3proxy@3proxy.ru>
+   (c) 2002-2021 by Vladimir Dubrovin <3proxy@3proxy.org>
 
    please read License Agreement
 
@@ -31,6 +31,9 @@ extern "C" {
 #include <arpa/inet.h>
 #include <netdb.h>
 #include <pthread.h>
+#ifdef IP_BOUND_IF
+#include <net/if.h>
+#endif
 #define SASIZETYPE socklen_t
 #define SOCKET int
 #define INVALID_SOCKET  (-1)
@@ -165,10 +168,10 @@ typedef enum {
 
 
 typedef enum {
-	S_NOSERVICE,
+	S_NOSERVICE = 0,
 	S_PROXY,
 	S_TCPPM,
-	S_POP3P,
+	S_POP3P = 3,
 	S_SOCKS4 = 4,	/* =4 */
 	S_SOCKS5 = 5,	/* =5 */
 	S_UDPPM,
@@ -180,7 +183,9 @@ typedef enum {
 	S_SMTPP,
 	S_REVLI,
 	S_REVCO,
-	S_ZOMBIE
+	S_ZOMBIE,
+	S_AUTO,
+	S_TLSPR
 }PROXYSERVICE;
 
 struct clientparam;
@@ -275,7 +280,8 @@ typedef enum {
 	R_SOCKS4B,
 	R_SOCKS5B,
 	R_ADMIN,
-	R_EXTIP
+	R_EXTIP,
+	R_TLS
 } REDIRTYPE;
 
 struct chain {
@@ -414,7 +420,51 @@ struct filterp {
 
 #define MAX_FILTERS 16
 
+
+struct sockfuncs {
+	void * state;
+	void (*freefunc)(void* state);
+#ifdef _WIN32
+	SOCKET (WINAPI *_socket)(void* state, int domain, int type, int protocol);
+	SOCKET (WINAPI *_accept)(void* state, SOCKET s, struct sockaddr * addr, int * addrlen);
+	int (WINAPI *_bind)(void* state, SOCKET s, const struct sockaddr *addr, int addrlen);
+	int (WINAPI *_listen)(void* state, SOCKET s, int backlog);
+	int (WINAPI *_connect)(void* state, SOCKET s, const struct sockaddr *name, int namelen);
+	int (WINAPI *_getpeername)(void* state, SOCKET s, struct sockaddr * name, int * namelen);
+	int (WINAPI *_getsockname)(void* state, SOCKET s, struct sockaddr * name, int * namelen);
+   	int (WINAPI *_getsockopt)(void* state, SOCKET s, int level, int optname, char * optval, int * optlen);
+	int (WINAPI *_setsockopt)(void* state, SOCKET s, int level, int optname, const char *optval, int optlen);
+	int (WINAPI *_poll)(void* state, struct pollfd *fds, unsigned int nfds, int timeout);
+	int (WINAPI *_send)(void* state, SOCKET s, const char *msg, int len, int flags);
+	int  (WINAPI *_sendto)(void* state, SOCKET s, const char *msg, int len, int flags, const struct sockaddr *to, int tolen);
+	int  (WINAPI *_recv)(void* state, SOCKET s, char *buf, int len, int flags);
+	int  (WINAPI *_recvfrom)(void* state, SOCKET s, char * buf, int len, int flags, struct sockaddr * from, int * fromlen);
+	int (WINAPI *_shutdown)(void* state, SOCKET s, int how);
+	int (WINAPI *_closesocket)(void* state, SOCKET s);
+#else
+	SOCKET (*_socket)(void* state, int domain, int type, int protocol);
+	SOCKET (*_accept)(void* state, SOCKET s, struct sockaddr * addr, socklen_t * addrlen);
+	int (*_bind)(void* state, SOCKET s, const struct sockaddr *addr, socklen_t addrlen);
+	int (*_listen)(void* state, SOCKET s, int backlog);
+	int (*_connect)(void* state, SOCKET s, const struct sockaddr *name, socklen_t namelen);
+	int (*_getpeername)(void* state, SOCKET s, struct sockaddr * name, socklen_t * namelen);
+	int (*_getsockname)(void* state, SOCKET s, struct sockaddr * name, socklen_t * namelen);
+   	int (*_getsockopt)(void* state, SOCKET s, int level, int optname, void * optval, socklen_t * optlen);
+	int (*_setsockopt)(void* state, int s, int level, int optname, const void *optval, socklen_t optlen);
+	int (*_poll)(void* state, struct pollfd *fds, nfds_t nfds, int timeout);
+	ssize_t (*_send)(void* state, SOCKET s, const void *msg, size_t len, int flags);
+	ssize_t (*_sendto)(void* state, SOCKET s, const void *msg, size_t len, int flags, const struct sockaddr *to, SASIZETYPE tolen);
+	ssize_t (*_recv)(void* state, SOCKET s, void *buf, size_t len, int flags);
+	ssize_t (*_recvfrom)(void* state, SOCKET s, void * buf, size_t len, int flags, struct sockaddr * from, SASIZETYPE * fromlen);
+	int (*_shutdown)(void* state, SOCKET s, int how);
+	int (*_closesocket)(void* state, SOCKET s);
+#endif
+};
+
+extern struct sockfuncs so;
+
 struct srvparam {
+	struct sockfuncs so;
 	struct srvparam *next;
 	struct srvparam *prev;
 	struct clientparam *child;
@@ -425,6 +475,7 @@ struct srvparam {
 	SOCKET srvsock, cbsock;
 	int childcount;
 	int maxchild;
+	int backlog;
 	int paused, version;
 	int singlepacket;
 	int usentlm;
@@ -437,6 +488,8 @@ struct srvparam {
 	int noforce;
 	int anonymous;
 	int clisockopts, srvsockopts, lissockopts, cbcsockopts, cbssockopts;
+	int gracetraf, gracenum, gracedelay;
+	int requirecert;
 #ifdef WITHSPLICE
 	int usesplice;
 #endif
@@ -456,7 +509,7 @@ struct srvparam {
 	struct pollfd fds;
 	FILE *stdlog;
 	unsigned char * target;
-#ifdef SO_BINDTODEVICE
+#if defined SO_BINDTODEVICE || defined IP_BOUND_IF
 	char * ibindtodevice;
 	char * obindtodevice;
 #endif
@@ -476,6 +529,7 @@ struct srvparam {
 struct clientparam {
 	struct clientparam	*next,
 				*prev;
+	void * sostate;
 	struct srvparam *srv;
 	REDIRECTFUNC redirectfunc;
 	BANDLIMFUNC bandlimfunc;
@@ -517,7 +571,8 @@ struct clientparam {
 		transparent,
 		chunked,
 		paused,
-		version;
+		version,
+		connlim;
 
 	unsigned char 	*hostname,
 			*username,
@@ -573,9 +628,10 @@ struct extparam {
 	struct srvparam *services;
 	int stacksize,
 		threadinit, counterd, haveerror, rotate, paused, archiverc,
-		demon, maxchild, needreload, timetoexit, version, noforce, bandlimver, parentretries;
+		demon, maxchild, backlog, needreload, timetoexit, version, noforce, bandlimver, parentretries;
 	int authcachetype, authcachetime;
 	int filtermaxsize;
+	int gracetraf, gracenum, gracedelay;
 	unsigned char *logname, **archiver;
 	ROTATION logtype, countertype;
 	char * counterfile;
@@ -682,46 +738,6 @@ struct hashtable {
 
 extern struct hashtable dns_table;
 extern struct hashtable dns6_table;
-
-struct sockfuncs {
-#ifdef _WIN32
-	SOCKET (WINAPI *_socket)(int domain, int type, int protocol);
-	SOCKET (WINAPI *_accept)(SOCKET s, struct sockaddr * addr, int * addrlen);
-	int (WINAPI *_bind)(SOCKET s, const struct sockaddr *addr, int addrlen);
-	int (WINAPI *_listen)(SOCKET s, int backlog);
-	int (WINAPI *_connect)(SOCKET s, const struct sockaddr *name, int namelen);
-	int (WINAPI *_getpeername)(SOCKET s, struct sockaddr * name, int * namelen);
-	int (WINAPI *_getsockname)(SOCKET s, struct sockaddr * name, int * namelen);
-   	int (WINAPI *_getsockopt)(SOCKET s, int level, int optname, char * optval, int * optlen);
-	int (WINAPI *_setsockopt)(SOCKET s, int level, int optname, const char *optval, int optlen);
-	int (WINAPI *_poll)(struct pollfd *fds, unsigned int nfds, int timeout);
-	int (WINAPI *_send)(SOCKET s, const char *msg, int len, int flags);
-	int  (WINAPI *_sendto)(SOCKET s, const char *msg, int len, int flags, const struct sockaddr *to, int tolen);
-	int  (WINAPI *_recv)(SOCKET s, char *buf, int len, int flags);
-	int  (WINAPI *_recvfrom)(SOCKET s, char * buf, int len, int flags, struct sockaddr * from, int * fromlen);
-	int (WINAPI *_shutdown)(SOCKET s, int how);
-	int (WINAPI *_closesocket)(SOCKET s);
-#else
-	SOCKET (*_socket)(int domain, int type, int protocol);
-	SOCKET (*_accept)(SOCKET s, struct sockaddr * addr, socklen_t * addrlen);
-	int (*_bind)(SOCKET s, const struct sockaddr *addr, socklen_t addrlen);
-	int (*_listen)(SOCKET s, int backlog);
-	int (*_connect)(SOCKET s, const struct sockaddr *name, socklen_t namelen);
-	int (*_getpeername)(SOCKET s, struct sockaddr * name, socklen_t * namelen);
-	int (*_getsockname)(SOCKET s, struct sockaddr * name, socklen_t * namelen);
-   	int (*_getsockopt)(SOCKET s, int level, int optname, void * optval, socklen_t * optlen);
-	int (*_setsockopt)(int s, int level, int optname, const void *optval, socklen_t optlen);
-	int (*_poll)(struct pollfd *fds, unsigned int nfds, int timeout);
-	size_t (*_send)(SOCKET s, const void *msg, size_t len, int flags);
-	size_t (*_sendto)(SOCKET s, const void *msg, size_t len, int flags, const struct sockaddr *to, SASIZETYPE tolen);
-	size_t (*_recv)(SOCKET s, void *buf, size_t len, int flags);
-	size_t (*_recvfrom)(SOCKET s, void * buf, size_t len, int flags, struct sockaddr * from, SASIZETYPE * fromlen);
-	int (*_shutdown)(SOCKET s, int how);
-	int (*_closesocket)(SOCKET s);
-#endif
-};
-
-extern struct sockfuncs so;
 struct pluginlink {
 	struct symbol symbols;
 	struct extparam *conf;
@@ -730,9 +746,9 @@ struct pluginlink {
 	struct auth *authfuncs;
 	struct commands * commandhandlers;
 	void * (*findbyname)(const char *name);
-	int (*socksend)(SOCKET sock, unsigned char * buf, int bufsize, int to);
-	int (*socksendto)(SOCKET sock, struct sockaddr * sin, unsigned char * buf, int bufsize, int to);
-	int (*sockrecvfrom)(SOCKET sock, struct sockaddr * sin, unsigned char * buf, int bufsize, int to);
+	int (*socksend)(struct clientparam *param, SOCKET sock, unsigned char * buf, int bufsize, int to);
+	int (*socksendto)(struct clientparam *param, SOCKET sock, struct sockaddr * sin, unsigned char * buf, int bufsize, int to);
+	int (*sockrecvfrom)(struct clientparam *param, SOCKET sock, struct sockaddr * sin, unsigned char * buf, int bufsize, int to);
 	int (*sockgetcharcli)(struct clientparam * param, int timeosec, int timeousec);
 	int (*sockgetcharsrv)(struct clientparam * param, int timeosec, int timeousec);
 	int (*sockgetlinebuf)(struct clientparam * param, DIRECTION which, unsigned char * buf, int bufsize, int delim, int to);
@@ -745,8 +761,8 @@ struct pluginlink {
 	int (*ACLMatches)(struct ace* acentry, struct clientparam * param);
 	int (*alwaysauth)(struct clientparam * param);
 	int (*checkACL)(struct clientparam * param);
-	void (*nametohash)(const unsigned char * name, unsigned char *hash);
-	unsigned (*hashindex)(const unsigned char* hash);
+	void (*nametohash)(const unsigned char * name, unsigned char *hash, unsigned char *rnd);
+	unsigned (*hashindex)(struct hashtable *ht, const unsigned char* hash);
 	unsigned char* (*en64)(const unsigned char *in, unsigned char *out, int inlen);
 	int (*de64)(const unsigned char *in, unsigned char *out, int maxlen);
 	void (*tohex)(unsigned char *in, unsigned char *out, int len);
